@@ -1,12 +1,12 @@
 #![allow(clippy::disallowed_names)]
-#[cfg(all(unix, not(feature = "tokio")))]
+use std::collections::HashMap;
+#[cfg(all(unix, not(feature = "tokio"), feature = "p2p"))]
 use std::os::unix::net::UnixStream;
-use std::{collections::HashMap, convert::TryInto};
-#[cfg(all(unix, feature = "tokio"))]
+#[cfg(all(unix, feature = "tokio", feature = "p2p"))]
 use tokio::net::UnixStream;
 
 use event_listener::Event;
-use futures_util::StreamExt;
+use futures_util::{StreamExt, TryStreamExt};
 use ntest::timeout;
 use serde::{Deserialize, Serialize};
 use test_log::test;
@@ -15,13 +15,18 @@ use tracing::{debug, instrument};
 use zbus::{
     block_on,
     fdo::{ObjectManager, ObjectManagerProxy},
-    DBusError,
+    message,
+    object_server::ResponseDispatchNotifier,
+    DBusError, Error, Message, MessageStream,
 };
-use zvariant::{DeserializeDict, OwnedValue, SerializeDict, Str, Type, Value};
+use zvariant::{DeserializeDict, Optional, OwnedValue, SerializeDict, Str, Type, Value};
 
 use zbus::{
-    dbus_interface, dbus_proxy, CacheProperties, Connection, ConnectionBuilder, InterfaceRef,
-    MessageHeader, MessageType, ObjectServer, SignalContext,
+    connection, interface,
+    message::Header,
+    object_server::{InterfaceRef, SignalEmitter},
+    proxy::CacheProperties,
+    Connection, ObjectServer,
 };
 
 #[derive(Debug, Deserialize, Serialize, Type)]
@@ -46,71 +51,6 @@ pub struct RefType<'a> {
     field1: Str<'a>,
 }
 
-#[dbus_proxy(assume_defaults = true, gen_blocking = true)]
-trait MyIface {
-    fn ping(&self) -> zbus::Result<u32>;
-
-    fn quit(&self) -> zbus::Result<()>;
-
-    fn test_header(&self) -> zbus::Result<()>;
-
-    fn test_error(&self) -> zbus::Result<()>;
-
-    fn test_single_struct_arg(&self, arg: ArgStructTest) -> zbus::Result<()>;
-
-    fn test_single_struct_ret(&self) -> zbus::Result<ArgStructTest>;
-
-    fn test_multi_ret(&self) -> zbus::Result<(i32, String)>;
-
-    fn test_hashmap_return(&self) -> zbus::Result<HashMap<String, String>>;
-
-    fn create_obj(&self, key: &str) -> zbus::Result<()>;
-
-    fn destroy_obj(&self, key: &str) -> zbus::Result<()>;
-
-    #[dbus_proxy(property)]
-    fn count(&self) -> zbus::Result<u32>;
-
-    #[dbus_proxy(property)]
-    fn set_count(&self, count: u32) -> zbus::Result<()>;
-
-    #[dbus_proxy(property)]
-    fn hash_map(&self) -> zbus::Result<HashMap<String, String>>;
-
-    #[dbus_proxy(property)]
-    fn address_data(&self) -> zbus::Result<IP4Adress>;
-
-    #[dbus_proxy(property)]
-    fn set_address_data(&self, addr: IP4Adress) -> zbus::Result<()>;
-
-    #[dbus_proxy(property)]
-    fn address_data2(&self) -> zbus::Result<IP4Adress>;
-
-    #[dbus_proxy(property)]
-    fn str_prop(&self) -> zbus::Result<String>;
-
-    #[dbus_proxy(property)]
-    fn set_str_prop(&self, str_prop: &str) -> zbus::Result<()>;
-
-    #[dbus_proxy(property)]
-    fn ref_type(&self) -> zbus::Result<RefType<'_>>;
-
-    #[dbus_proxy(property)]
-    fn set_ref_type(&self, ref_type: RefType<'_>) -> zbus::Result<()>;
-
-    #[dbus_proxy(property)]
-    fn fail_property(&self) -> zbus::Result<u32>;
-
-    #[dbus_proxy(no_reply)]
-    fn test_no_reply(&self) -> zbus::Result<()>;
-
-    #[dbus_proxy(no_autostart)]
-    fn test_no_autostart(&self) -> zbus::Result<()>;
-
-    #[dbus_proxy(allow_interactive_auth)]
-    fn test_interactive_auth(&self) -> zbus::Result<()>;
-}
-
 #[derive(Debug, Clone)]
 enum NextAction {
     Quit,
@@ -119,33 +59,50 @@ enum NextAction {
 }
 
 #[derive(Debug)]
-struct MyIfaceImpl {
+struct MyIface {
     next_tx: Sender<NextAction>,
     count: u32,
+    emits_changed_default: u32,
+    emits_changed_true: u32,
+    emits_changed_invalidates: u32,
+    emits_changed_const: u32,
+    emits_changed_false: u32,
 }
 
-impl MyIfaceImpl {
+impl MyIface {
     fn new(next_tx: Sender<NextAction>) -> Self {
-        Self { next_tx, count: 0 }
+        Self {
+            next_tx,
+            count: 0,
+            emits_changed_default: 0,
+            emits_changed_true: 0,
+            emits_changed_invalidates: 0,
+            emits_changed_const: 0,
+            emits_changed_false: 0,
+        }
     }
 }
 
 /// Custom D-Bus error type.
-#[derive(Debug, DBusError)]
-#[dbus_error(prefix = "org.freedesktop.MyIface.Error")]
+#[derive(Debug, DBusError, PartialEq)]
+#[zbus(prefix = "org.freedesktop.MyIface.Error")]
 enum MyIfaceError {
     SomethingWentWrong(String),
-    #[dbus_error(zbus_error)]
+    #[zbus(error)]
     ZBus(zbus::Error),
 }
 
-#[dbus_interface(interface = "org.freedesktop.MyIface")]
-impl MyIfaceImpl {
+#[interface(
+    interface = "org.freedesktop.MyIface",
+    proxy(gen_blocking = true, assume_defaults = true, visibility = "pub(self)")
+)]
+impl MyIface {
     #[instrument]
-    async fn ping(&mut self, #[zbus(signal_context)] ctxt: SignalContext<'_>) -> u32 {
+    async fn ping(&mut self, #[zbus(signal_emitter)] emitter: SignalEmitter<'_>) -> u32 {
         self.count += 1;
         if self.count % 3 == 0 {
-            MyIfaceImpl::alert_count(&ctxt, self.count)
+            emitter
+                .alert_count(self.count)
                 .await
                 .expect("Failed to emit signal");
             debug!("emitted `AlertCount` signal.");
@@ -162,10 +119,10 @@ impl MyIfaceImpl {
     }
 
     #[instrument]
-    fn test_header(&self, #[zbus(header)] header: MessageHeader<'_>) {
+    fn test_header(&self, #[zbus(header)] header: Header<'_>) {
         debug!("`TestHeader` called.");
-        assert_eq!(header.message_type().unwrap(), MessageType::MethodCall);
-        assert_eq!(header.member().unwrap().unwrap(), "TestHeader");
+        assert_eq!(header.message_type(), message::Type::MethodCall);
+        assert_eq!(header.member().unwrap(), "TestHeader");
     }
 
     #[instrument]
@@ -184,10 +141,10 @@ impl MyIfaceImpl {
     fn test_single_struct_arg(
         &self,
         arg: ArgStructTest,
-        #[zbus(header)] header: MessageHeader<'_>,
+        #[zbus(header)] header: Header<'_>,
     ) -> zbus::fdo::Result<()> {
         debug!("`TestSingleStructArg` called.");
-        assert_eq!(header.signature()?.unwrap(), "(is)");
+        assert_eq!(header.signature(), "(is)");
         assert_eq!(arg.foo, 1);
         assert_eq!(arg.bar, "TestString");
 
@@ -204,12 +161,38 @@ impl MyIfaceImpl {
     }
 
     #[instrument]
-    #[dbus_interface(out_args("foo", "bar"))]
+    #[zbus(out_args("foo", "bar"))]
     fn test_multi_ret(&self) -> zbus::fdo::Result<(i32, String)> {
         debug!("`TestMultiRet` called.");
         Ok((42, String::from("Meaning of life")))
     }
 
+    #[instrument]
+    fn test_response_notify(
+        &self,
+        #[zbus(connection)] conn: &Connection,
+        #[zbus(signal_emitter)] emitter: SignalEmitter<'_>,
+    ) -> zbus::fdo::Result<ResponseDispatchNotifier<String>> {
+        debug!("`TestResponseNotify` called.");
+        let (response, listener) = ResponseDispatchNotifier::new(String::from("Meaning of life"));
+        let emitter = emitter.to_owned();
+        conn.executor()
+            .spawn(
+                async move {
+                    listener.await;
+
+                    Self::test_response_notified(emitter).await.unwrap();
+                },
+                "TestResponseNotify",
+            )
+            .detach();
+
+        Ok(response)
+    }
+
+    #[zbus(signal)]
+    async fn test_response_notified(emitter: SignalEmitter<'_>) -> zbus::Result<()>;
+
     #[instrument]
     async fn test_hashmap_return(&self) -> zbus::fdo::Result<HashMap<String, String>> {
         debug!("`TestHashmapReturn` called.");
@@ -221,9 +204,12 @@ impl MyIfaceImpl {
     }
 
     #[instrument]
-    async fn create_obj(&self, key: String) {
+    async fn create_obj(&self, key: &str) {
         debug!("`CreateObj` called.");
-        self.next_tx.send(NextAction::CreateObj(key)).await.unwrap();
+        self.next_tx
+            .send(NextAction::CreateObj(key.into()))
+            .await
+            .unwrap();
     }
 
     #[instrument]
@@ -236,23 +222,30 @@ impl MyIfaceImpl {
         object_server
             .at(
                 format!("/zbus/test/{key}"),
-                MyIfaceImpl::new(self.next_tx.clone()),
+                MyIface::new(self.next_tx.clone()),
             )
             .await
             .unwrap();
     }
 
     #[instrument]
-    async fn destroy_obj(&self, key: String) {
+    async fn destroy_obj(&self, key: &str) {
         debug!("`DestroyObj` called.");
         self.next_tx
-            .send(NextAction::DestroyObj(key))
+            .send(NextAction::DestroyObj(key.into()))
             .await
             .unwrap();
     }
 
+    #[cfg(feature = "option-as-array")]
     #[instrument]
-    #[dbus_interface(property)]
+    async fn optional_args(&self, arg: Option<&str>) -> zbus::fdo::Result<Option<String>> {
+        debug!("`OptionalArgs` called.");
+        Ok(arg.map(|s| format!("Hello {}", s)))
+    }
+
+    #[instrument]
+    #[zbus(property)]
     fn set_count(&mut self, val: u32) -> zbus::fdo::Result<()> {
         debug!("`Count` setter called.");
         if val == 42 {
@@ -263,21 +256,53 @@ impl MyIfaceImpl {
     }
 
     #[instrument]
-    #[dbus_interface(property)]
+    #[zbus(property)]
     fn count(&self) -> u32 {
         debug!("`Count` getter called.");
         self.count
     }
 
     #[instrument]
-    #[dbus_interface(property)]
+    #[zbus(property)]
+    fn test_header_prop(
+        &self,
+        #[zbus(header)] header: Option<Header<'_>>,
+        #[zbus(connection)] connection: &Connection,
+        #[zbus(object_server)] object_server: &ObjectServer,
+        #[zbus(signal_emitter)] emitter: SignalEmitter<'_>,
+    ) -> bool {
+        debug!(
+            "`TestHeaderProp` getter called, header: {:?}, connection: {:?}, object_server: {:?}, emitter: {:?}",
+            header, connection, object_server, emitter
+        );
+        header.is_some()
+    }
+
+    #[instrument]
+    #[zbus(property)]
+    fn set_test_header_prop(
+        &self,
+        value: bool,
+        #[zbus(header)] header: Option<Header<'_>>,
+        #[zbus(connection)] connection: &Connection,
+        #[zbus(object_server)] object_server: &ObjectServer,
+        #[zbus(signal_emitter)] emitter: SignalEmitter<'_>,
+    ) {
+        debug!("`TestHeaderProp` setter called, value: {}, header: {:?}, connection: {:?}, object_server: {:?}, emitter: {:?}",
+            value, header, connection, object_server, emitter
+        );
+        assert!(header.is_some());
+    }
+
+    #[instrument]
+    #[zbus(property)]
     async fn hash_map(&self) -> HashMap<String, String> {
         debug!("`HashMap` getter called.");
         self.test_hashmap_return().await.unwrap()
     }
 
     #[instrument]
-    #[dbus_interface(property)]
+    #[zbus(property)]
     async fn fail_property(&self) -> zbus::fdo::Result<u32> {
         Err(zbus::fdo::Error::UnknownProperty(
             "FailProperty".to_string(),
@@ -285,7 +310,14 @@ impl MyIfaceImpl {
     }
 
     #[instrument]
-    #[dbus_interface(property)]
+    #[zbus(property)]
+    fn optional_property(&self) -> Optional<u32> {
+        debug!("`OptionalAsProp` getter called.");
+        Some(42).into()
+    }
+
+    #[instrument]
+    #[zbus(property)]
     fn address_data(&self) -> IP4Adress {
         debug!("`AddressData` getter called.");
         IP4Adress {
@@ -295,7 +327,7 @@ impl MyIfaceImpl {
     }
 
     #[instrument]
-    #[dbus_interface(property)]
+    #[zbus(property)]
     fn set_address_data(&self, addr: IP4Adress) {
         debug!("`AddressData` setter called with {:?}", addr);
     }
@@ -303,30 +335,33 @@ impl MyIfaceImpl {
     // On the bus, this should return the same value as address_data above. We want to test if
     // this works both ways.
     #[instrument]
-    #[dbus_interface(property)]
+    #[zbus(property)]
     fn address_data2(&self) -> HashMap<String, OwnedValue> {
         debug!("`AddressData2` getter called.");
         let mut map = HashMap::new();
-        map.insert("address".into(), Value::from("127.0.0.1").into());
+        map.insert(
+            "address".into(),
+            Value::from("127.0.0.1").try_into().unwrap(),
+        );
         map.insert("prefix".into(), 1234u32.into());
 
         map
     }
 
     #[instrument]
-    #[dbus_interface(property)]
+    #[zbus(property)]
     fn str_prop(&self) -> String {
         "Hello".to_string()
     }
 
     #[instrument]
-    #[dbus_interface(property)]
+    #[zbus(property)]
     fn set_str_prop(&self, str_prop: &str) {
         debug!("`SetStrRef` called with {:?}", str_prop);
     }
 
     #[instrument]
-    #[dbus_interface(property)]
+    #[zbus(property)]
     fn ref_prop(&self) -> RefType<'_> {
         RefType {
             field1: "Hello".into(),
@@ -334,52 +369,121 @@ impl MyIfaceImpl {
     }
 
     #[instrument]
-    #[dbus_interface(property)]
+    #[zbus(property)]
     fn set_ref_prop(&self, ref_type: RefType<'_>) {
         debug!("`SetRefType` called with {:?}", ref_type);
     }
 
     #[instrument]
-    fn test_no_reply(&self, #[zbus(header)] header: MessageHeader<'_>) {
+    #[zbus(proxy(no_reply))]
+    fn test_no_reply(&self, #[zbus(header)] header: Header<'_>) {
         debug!("`TestNoReply` called");
-        assert_eq!(
-            header.message_type().unwrap(),
-            zbus::MessageType::MethodCall
-        );
+        assert_eq!(header.message_type(), zbus::message::Type::MethodCall);
         assert!(header
             .primary()
             .flags()
-            .contains(zbus::MessageFlags::NoReplyExpected));
+            .contains(zbus::message::Flags::NoReplyExpected));
     }
 
     #[instrument]
-    fn test_no_autostart(&self, #[zbus(header)] header: MessageHeader<'_>) {
+    #[zbus(proxy(no_autostart))]
+    fn test_no_autostart(&self, #[zbus(header)] header: Header<'_>) {
         debug!("`TestNoAutostart` called");
-        assert_eq!(
-            header.message_type().unwrap(),
-            zbus::MessageType::MethodCall
-        );
+        assert_eq!(header.message_type(), zbus::message::Type::MethodCall);
         assert!(header
             .primary()
             .flags()
-            .contains(zbus::MessageFlags::NoAutoStart));
+            .contains(zbus::message::Flags::NoAutoStart));
     }
 
     #[instrument]
-    fn test_interactive_auth(&self, #[zbus(header)] header: MessageHeader<'_>) {
+    #[zbus(proxy(allow_interactive_auth))]
+    fn test_interactive_auth(&self, #[zbus(header)] header: Header<'_>) {
         debug!("`TestInteractiveAuth` called");
-        assert_eq!(
-            header.message_type().unwrap(),
-            zbus::MessageType::MethodCall
-        );
+        assert_eq!(header.message_type(), zbus::message::Type::MethodCall);
         assert!(header
             .primary()
             .flags()
-            .contains(zbus::MessageFlags::AllowInteractiveAuth));
+            .contains(zbus::message::Flags::AllowInteractiveAuth));
     }
 
-    #[dbus_interface(signal)]
-    async fn alert_count(ctxt: &SignalContext<'_>, val: u32) -> zbus::Result<()>;
+    #[zbus(signal)]
+    async fn alert_count(emitter: &SignalEmitter<'_>, val: u32) -> zbus::Result<()>;
+
+    #[instrument]
+    #[zbus(property)]
+    fn emits_changed_default(&self) -> u32 {
+        debug!("`EmitsChangedDefault` getter called.");
+        self.emits_changed_default
+    }
+
+    #[instrument]
+    #[zbus(property)]
+    fn set_emits_changed_default(&mut self, val: u32) -> zbus::fdo::Result<()> {
+        debug!("`EmitsChangedDefault` setter called.");
+        self.emits_changed_default = val;
+        Ok(())
+    }
+
+    #[instrument]
+    #[zbus(property(emits_changed_signal = "true"))]
+    fn emits_changed_true(&self) -> u32 {
+        debug!("`EmitsChangedTrue` getter called.");
+        self.emits_changed_true
+    }
+
+    #[instrument]
+    #[zbus(property)]
+    fn set_emits_changed_true(&mut self, val: u32) -> zbus::fdo::Result<()> {
+        debug!("`EmitsChangedTrue` setter called.");
+        self.emits_changed_true = val;
+        Ok(())
+    }
+
+    #[instrument]
+    #[zbus(property(emits_changed_signal = "invalidates"))]
+    fn emits_changed_invalidates(&self) -> u32 {
+        debug!("`EmitsChangedInvalidates` getter called.");
+        self.emits_changed_invalidates
+    }
+
+    #[instrument]
+    #[zbus(property)]
+    fn set_emits_changed_invalidates(&mut self, val: u32) -> zbus::fdo::Result<()> {
+        debug!("`EmitsChangedInvalidates` setter called.");
+        self.emits_changed_invalidates = val;
+        Ok(())
+    }
+
+    #[instrument]
+    #[zbus(property(emits_changed_signal = "const"))]
+    fn emits_changed_const(&self) -> u32 {
+        debug!("`EmitsChangedConst` getter called.");
+        self.emits_changed_const
+    }
+
+    #[instrument]
+    #[zbus(property)]
+    fn set_emits_changed_const(&mut self, val: u32) -> zbus::fdo::Result<()> {
+        debug!("`EmitsChangedConst` setter called.");
+        self.emits_changed_const = val;
+        Ok(())
+    }
+
+    #[instrument]
+    #[zbus(property(emits_changed_signal = "false"))]
+    fn emits_changed_false(&self) -> u32 {
+        debug!("`EmitsChangedFalse` getter called.");
+        self.emits_changed_false
+    }
+
+    #[instrument]
+    #[zbus(property)]
+    fn set_emits_changed_false(&mut self, val: u32) -> zbus::fdo::Result<()> {
+        debug!("`EmitsChangedFalse` setter called.");
+        self.emits_changed_false = val;
+        Ok(())
+    }
 }
 
 fn check_hash_map(map: HashMap<String, String>) {
@@ -397,9 +501,65 @@ fn check_ipv4_address(address: IP4Adress) {
     );
 }
 
+fn check_ipv4_address_hashmap(address: HashMap<String, OwnedValue>) {
+    assert_eq!(**address.get("address").unwrap(), Value::from("127.0.0.1"));
+    assert_eq!(**address.get("prefix").unwrap(), Value::from(1234u32));
+}
+
 #[instrument]
 async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
     debug!("client side starting..");
+    // Use low-level API for `TestResponseNotify` because we need to ensure that the signal is
+    // always received after the response.
+    let mut stream = MessageStream::from(&conn);
+    let method = Message::method_call("/org/freedesktop/MyService", "TestResponseNotify")?
+        .interface("org.freedesktop.MyIface")?
+        .destination("org.freedesktop.MyService")?
+        .build(&())?;
+    let serial = method.primary_header().serial_num();
+    conn.send(&method).await?;
+    let mut method_returned = false;
+    let mut signal_received = false;
+    while !method_returned && !signal_received {
+        let msg = stream.try_next().await?.unwrap();
+
+        let hdr = msg.header();
+        if hdr.message_type() == message::Type::MethodReturn && hdr.reply_serial() == Some(serial) {
+            assert!(!signal_received);
+            method_returned = true;
+        } else if hdr.message_type() == message::Type::Signal
+            && hdr.interface().unwrap() == "org.freedesktop.MyService"
+            && hdr.member().unwrap() == "TestResponseNotified"
+        {
+            assert!(method_returned);
+            signal_received = true;
+        }
+    }
+    drop(stream);
+
+    let root_introspect_proxy = zbus::fdo::IntrospectableProxy::builder(&conn)
+        .destination("org.freedesktop.MyService")?
+        .path("/")?
+        .build()
+        .await?;
+    debug!("Created: {:?}", root_introspect_proxy);
+
+    let root_xml = root_introspect_proxy.introspect().await?;
+    let root_node = zbus_xml::Node::from_reader(root_xml.as_bytes())
+        .map_err(|e| Error::Failure(e.to_string()))?;
+    let mut node = &root_node;
+    for name in ["org", "freedesktop", "MyService"] {
+        node = node
+            .nodes()
+            .iter()
+            .find(|&n| n.name().is_some_and(|n| n == name))
+            .expect("Child node not exist");
+    }
+    assert!(node
+        .interfaces()
+        .iter()
+        .any(|i| i.name() == "org.freedesktop.MyIface"));
+
     let proxy = MyIfaceProxy::builder(&conn)
         .destination("org.freedesktop.MyService")?
         .path("/org/freedesktop/MyService")?
@@ -408,6 +568,16 @@ async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
         .build()
         .await?;
     debug!("Created: {:?}", proxy);
+
+    // First let's call a non-existent method. It should immediately fail.
+    // There was a regression where object server would just not reply in this case:
+    // https://github.com/dbus2/zbus/issues/905
+    assert!(proxy
+        .inner()
+        .call::<_, _, ()>("NonExistantMethod", &())
+        .await
+        .is_err());
+
     let props_proxy = zbus::fdo::PropertiesProxy::builder(&conn)
         .destination("org.freedesktop.MyService")?
         .path("/org/freedesktop/MyService")?
@@ -432,6 +602,8 @@ async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
     drop(props_changed_stream);
 
     proxy.ping().await?;
+    proxy.set_test_header_prop(true).await?;
+    assert!(proxy.test_header_prop().await?);
     assert_eq!(proxy.count().await?, 1);
     assert_eq!(proxy.cached_count()?, None);
 
@@ -442,6 +614,13 @@ async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
             bar: "TestString".into(),
         })
         .await?;
+
+    proxy.test_error().await.unwrap_err();
+    assert_eq!(
+        proxy.test_custom_error().await.unwrap_err(),
+        MyIfaceError::SomethingWentWrong("oops".to_string())
+    );
+
     check_hash_map(proxy.test_hashmap_return().await?);
     check_hash_map(proxy.hash_map().await?);
     proxy
@@ -452,7 +631,7 @@ async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
         .await?;
     proxy.set_str_prop("This is an str ref").await?;
     check_ipv4_address(proxy.address_data().await?);
-    check_ipv4_address(proxy.address_data2().await?);
+    check_ipv4_address_hashmap(proxy.address_data2().await?);
 
     proxy.test_no_reply().await?;
     proxy.test_no_autostart().await?;
@@ -463,46 +642,40 @@ async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
         err.unwrap_err(),
         zbus::Error::FDO(Box::new(zbus::fdo::Error::UnknownProperty(
             "FailProperty".into()
-        )))
+        ))),
     );
 
-    #[cfg(any(feature = "xml", feature = "quick-xml"))]
-    {
-        let xml = proxy.introspect().await?;
-        debug!("Introspection: {}", xml);
-        #[cfg(all(feature = "xml", not(feature = "quick-xml")))]
-        let node = zbus::xml::Node::from_reader(xml.as_bytes())?;
-        #[cfg(feature = "quick-xml")]
-        let node = zbus::quick_xml::Node::from_reader(xml.as_bytes())?;
-        let ifaces = node.interfaces();
-        let iface = ifaces
-            .iter()
-            .find(|i| i.name() == "org.freedesktop.MyIface")
-            .unwrap();
-        let methods = iface.methods();
-        for method in methods {
-            if method.name() != "TestSingleStructRet" && method.name() != "TestMultiRet" {
-                continue;
-            }
-            let args = method.args();
-            #[cfg(all(feature = "xml", not(feature = "quick-xml")))]
-            let mut out_args = args.iter().filter(|a| a.direction().unwrap() == "out");
-            #[cfg(feature = "quick-xml")]
-            let mut out_args = args
-                .iter()
-                .filter(|a| a.direction().unwrap() == zbus::quick_xml::ArgDirection::Out);
+    assert_eq!(proxy.optional_property().await?, Some(42).into());
 
-            if method.name() == "TestSingleStructRet" {
-                assert_eq!(args.len(), 1);
-                assert_eq!(out_args.next().unwrap().ty(), "(is)");
-                assert!(out_args.next().is_none());
-            } else {
-                assert_eq!(args.len(), 2);
-                let foo = out_args.find(|a| a.name() == Some("foo")).unwrap();
-                assert_eq!(foo.ty(), "i");
-                let bar = out_args.find(|a| a.name() == Some("bar")).unwrap();
-                assert_eq!(bar.ty(), "s");
-            }
+    let xml = proxy.inner().introspect().await?;
+    debug!("Introspection: {}", xml);
+    let node =
+        zbus_xml::Node::from_reader(xml.as_bytes()).map_err(|e| Error::Failure(e.to_string()))?;
+    let ifaces = node.interfaces();
+    let iface = ifaces
+        .iter()
+        .find(|i| i.name() == "org.freedesktop.MyIface")
+        .unwrap();
+    let methods = iface.methods();
+    for method in methods {
+        if method.name() != "TestSingleStructRet" && method.name() != "TestMultiRet" {
+            continue;
+        }
+        let args = method.args();
+        let mut out_args = args
+            .iter()
+            .filter(|a| a.direction().unwrap() == zbus_xml::ArgDirection::Out);
+
+        if method.name() == "TestSingleStructRet" {
+            assert_eq!(args.len(), 1);
+            assert_eq!(out_args.next().unwrap().ty(), "(is)");
+            assert!(out_args.next().is_none());
+        } else {
+            assert_eq!(args.len(), 2);
+            let foo = out_args.find(|a| a.name() == Some("foo")).unwrap();
+            assert_eq!(foo.ty(), "i");
+            let bar = out_args.find(|a| a.name() == Some("bar")).unwrap();
+            assert_eq!(bar.ty(), "s");
         }
     }
     // build-time check to see if macro is doing the right thing.
@@ -534,6 +707,15 @@ async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
     )
     .await;
 
+    #[cfg(feature = "option-as-array")]
+    {
+        assert!(proxy.optional_args(None).await.unwrap().is_none());
+        assert_eq!(
+            proxy.optional_args(Some("🚌")).await.unwrap().unwrap(),
+            "Hello 🚌",
+        );
+    }
+
     assert_eq!(ifaces_added.args()?.object_path(), "/zbus/test/MyObj");
     let args = ifaces_added.args()?;
     let ifaces = args.interfaces_and_properties();
@@ -541,7 +723,7 @@ async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
     // TODO: Check if the properties are correct.
 
     // issue#207: interface panics on incorrect number of args.
-    assert!(proxy.call_method("CreateObj", &()).await.is_err());
+    assert!(proxy.inner().call_method("CreateObj", &()).await.is_err());
 
     let my_obj_proxy = MyIfaceProxy::builder(&conn)
         .destination("org.freedesktop.MyService")?
@@ -556,7 +738,7 @@ async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
     assert_eq!(my_obj_proxy.count().await?, 0);
     assert_eq!(my_obj_proxy.cached_count()?, Some(0));
     assert_eq!(
-        my_obj_proxy.cached_property_raw("Count").as_deref(),
+        my_obj_proxy.inner().cached_property_raw("Count").as_deref(),
         Some(&Value::from(0u32))
     );
     my_obj_proxy.ping().await?;
@@ -579,14 +761,15 @@ async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
 
     let args = ifaces_removed.args()?;
     assert_eq!(args.object_path(), "/zbus/test/MyObj");
-    assert_eq!(args.interfaces(), &["org.freedesktop.MyIface"]);
+    assert_eq!(args.interfaces().as_ref(), &["org.freedesktop.MyIface"]);
 
-    assert!(my_obj_proxy.introspect().await.is_err());
+    assert!(my_obj_proxy.inner().introspect().await.is_err());
     assert!(my_obj_proxy.ping().await.is_err());
 
     // Make sure methods modifying the ObjectServer can be called without
     // deadlocks.
     proxy
+        .inner()
         .call_method("CreateObjInside", &("CreatedInside"))
         .await?;
     let created_inside_proxy = MyIfaceProxy::builder(&conn)
@@ -597,6 +780,95 @@ async fn my_iface_test(conn: Connection, event: Event) -> zbus::Result<u32> {
     created_inside_proxy.ping().await?;
     proxy.destroy_obj("CreatedInside").await?;
 
+    // Test that interfaces emit signals when properties change
+    // according to their emits_changed_signal flags.
+    let mut props_changed = props_proxy.receive_properties_changed().await?;
+    let expected_property_value = 4;
+
+    proxy
+        .set_emits_changed_default(expected_property_value)
+        .await?;
+    let changed = props_changed.next().await.unwrap();
+    let expected_property_key = "EmitsChangedDefault";
+    let args = changed.args()?;
+    assert!(args.invalidated_properties().is_empty());
+
+    let changed_prop_key = *args.changed_properties().keys().next().unwrap();
+    assert_eq!(changed_prop_key, expected_property_key);
+
+    let changed_value = &args.changed_properties()[expected_property_key];
+    assert_eq!(
+        <&Value as TryInto<u32>>::try_into(changed_value).unwrap(),
+        expected_property_value
+    );
+    assert!(args.invalidated_properties().is_empty());
+
+    proxy
+        .set_emits_changed_true(expected_property_value)
+        .await?;
+    let changed = props_changed.next().await.unwrap();
+    let expected_property_key = "EmitsChangedTrue";
+    let args = changed.args()?;
+    assert!(args.invalidated_properties().is_empty());
+
+    let changed_prop_key = *args.changed_properties().keys().next().unwrap();
+    assert_eq!(changed_prop_key, expected_property_key);
+
+    let changed_value = &args.changed_properties()[expected_property_key];
+    assert_eq!(
+        <&Value as TryInto<u32>>::try_into(changed_value).unwrap(),
+        expected_property_value
+    );
+    assert!(args.invalidated_properties().is_empty());
+
+    proxy
+        .set_emits_changed_invalidates(expected_property_value)
+        .await?;
+    let changed = props_changed.next().await.unwrap();
+    let expected_property_key = "EmitsChangedInvalidates";
+    let args = changed.args()?;
+    assert_eq!(
+        args.invalidated_properties(),
+        &vec![expected_property_key.to_string()]
+    );
+    assert!(args.changed_properties().is_empty());
+
+    // First set a property for which we don't expect a signal
+    // then set a property for which we do (and we checked above
+    // that we receive it. The next item in the iter should correspond
+    // to the second property we set.
+    proxy
+        .set_emits_changed_const(expected_property_value)
+        .await?;
+    proxy
+        .set_emits_changed_true(expected_property_value)
+        .await?;
+    let changed = props_changed.next().await.unwrap();
+    let unexpected_property_key = "EmitsChangedConst";
+    let args = changed.args()?;
+    assert_ne!(
+        args.invalidated_properties(),
+        &vec![unexpected_property_key.to_string()]
+    );
+    assert!(!args.changed_properties().is_empty());
+    assert!(args.invalidated_properties().is_empty());
+
+    proxy
+        .set_emits_changed_false(expected_property_value)
+        .await?;
+    proxy
+        .set_emits_changed_true(expected_property_value)
+        .await?;
+    let changed = props_changed.next().await.unwrap();
+    let unexpected_property_key = "EmitsChangedFalse";
+    let args = changed.args()?;
+    assert_ne!(
+        args.invalidated_properties(),
+        &vec![unexpected_property_key.to_string()]
+    );
+    assert!(!args.changed_properties().is_empty());
+    assert!(args.invalidated_properties().is_empty());
+
     proxy.quit().await?;
     Ok(val)
 }
@@ -607,6 +879,7 @@ fn iface_and_proxy() {
     block_on(iface_and_proxy_(false));
 }
 
+#[cfg(feature = "p2p")]
 #[cfg(unix)]
 #[test]
 #[timeout(15000)]
@@ -615,18 +888,38 @@ fn iface_and_proxy_unix_p2p() {
 }
 
 #[instrument]
-async fn iface_and_proxy_(p2p: bool) {
+async fn iface_and_proxy_(#[allow(unused)] p2p: bool) {
     let event = event_listener::Event::new();
+    #[cfg(feature = "p2p")]
     let guid = zbus::Guid::generate();
 
+    let session_conns_build = || {
+        let service_conn_builder = connection::Builder::session()
+            .unwrap()
+            .name("org.freedesktop.MyService")
+            .unwrap()
+            .name("org.freedesktop.MyService.foo")
+            .unwrap()
+            .name("org.freedesktop.MyService.bar")
+            .unwrap()
+            .name("org.freedesktop.MyEmitsChangedSignalIface")
+            .unwrap();
+        let client_conn_builder = connection::Builder::session().unwrap();
+
+        (service_conn_builder, client_conn_builder)
+    };
+    #[cfg(feature = "p2p")]
     let (service_conn_builder, client_conn_builder) = if p2p {
         #[cfg(unix)]
         {
             let (p0, p1) = UnixStream::pair().unwrap();
 
             (
-                ConnectionBuilder::unix_stream(p0).server(&guid).p2p(),
-                ConnectionBuilder::unix_stream(p1).p2p(),
+                connection::Builder::unix_stream(p0)
+                    .server(guid)
+                    .unwrap()
+                    .p2p(),
+                connection::Builder::unix_stream(p1).p2p(),
             )
         }
 
@@ -640,8 +933,11 @@ async fn iface_and_proxy_(p2p: bool) {
                 let p0 = listener.incoming().next().unwrap().unwrap();
 
                 (
-                    ConnectionBuilder::tcp_stream(p0).server(&guid).p2p(),
-                    ConnectionBuilder::tcp_stream(p1).p2p(),
+                    connection::Builder::tcp_stream(p0)
+                        .server(guid)
+                        .unwrap()
+                        .p2p(),
+                    connection::Builder::tcp_stream(p1).p2p(),
                 )
             }
 
@@ -653,24 +949,20 @@ async fn iface_and_proxy_(p2p: bool) {
                 let p0 = listener.accept().await.unwrap().0;
 
                 (
-                    ConnectionBuilder::tcp_stream(p0).server(&guid).p2p(),
-                    ConnectionBuilder::tcp_stream(p1).p2p(),
+                    connection::Builder::tcp_stream(p0)
+                        .server(guid)
+                        .unwrap()
+                        .p2p(),
+                    connection::Builder::tcp_stream(p1).p2p(),
                 )
             }
         }
     } else {
-        let service_conn_builder = ConnectionBuilder::session()
-            .unwrap()
-            .name("org.freedesktop.MyService")
-            .unwrap()
-            .name("org.freedesktop.MyService.foo")
-            .unwrap()
-            .name("org.freedesktop.MyService.bar")
-            .unwrap();
-        let client_conn_builder = ConnectionBuilder::session().unwrap();
-
-        (service_conn_builder, client_conn_builder)
+        session_conns_build()
     };
+    #[cfg(not(feature = "p2p"))]
+    let (service_conn_builder, client_conn_builder) = session_conns_build();
+
     debug!(
         "Client connection builder created: {:?}",
         client_conn_builder
@@ -680,7 +972,7 @@ async fn iface_and_proxy_(p2p: bool) {
         service_conn_builder
     );
     let (next_tx, mut next_rx) = channel(64);
-    let iface = MyIfaceImpl::new(next_tx.clone());
+    let iface = MyIface::new(next_tx.clone());
     let service_conn_builder = service_conn_builder
         .serve_at("/org/freedesktop/MyService", iface)
         .unwrap()
@@ -703,7 +995,7 @@ async fn iface_and_proxy_(p2p: bool) {
     listen.await;
     debug!("Child task signaled it's ready.");
 
-    let iface: InterfaceRef<MyIfaceImpl> = service_conn
+    let iface: InterfaceRef<MyIface> = service_conn
         .object_server()
         .interface("/org/freedesktop/MyService")
         .await
@@ -711,15 +1003,13 @@ async fn iface_and_proxy_(p2p: bool) {
     iface
         .get()
         .await
-        .count_changed(iface.signal_context())
+        .count_changed(iface.signal_emitter())
         .await
         .unwrap();
     debug!("`PropertiesChanged` emitted for `Count` property.");
 
     loop {
-        MyIfaceImpl::alert_count(iface.signal_context(), 51)
-            .await
-            .unwrap();
+        iface.alert_count(51).await.unwrap();
         debug!("`AlertCount` signal emitted.");
 
         match next_rx.recv().await.unwrap() {
@@ -728,7 +1018,7 @@ async fn iface_and_proxy_(p2p: bool) {
                 let path = format!("/zbus/test/{key}");
                 service_conn
                     .object_server()
-                    .at(path.clone(), MyIfaceImpl::new(next_tx.clone()))
+                    .at(path.clone(), MyIface::new(next_tx.clone()))
                     .await
                     .unwrap();
                 debug!("Object `{path}` added.");
@@ -737,7 +1027,7 @@ async fn iface_and_proxy_(p2p: bool) {
                 let path = format!("/zbus/test/{key}");
                 service_conn
                     .object_server()
-                    .remove::<MyIfaceImpl, _>(path.clone())
+                    .remove::<MyIface, _>(path.clone())
                     .await
                     .unwrap();
                 debug!("Object `{path}` removed.");
